"""multiprocessing tools for multi core and multi node systems."""
import inspect
import copy
import multiprocessing as mp
import os
from os.path import join as pjoin
from pathlib import Path
import re
import shlex
import subprocess
import time
import warnings

import numpy as np
import pickle

#: SLURM parameters
slurm_options = {
    # The HPC user account
    "account": None, # required
    "job-name": None, # required
    "ntasks-per-node": None, # required, number of cores in each compute node
    "qos": None, # Quality of service
    "reservation": None, # Request particular resources
    "time": 60, # job minutes
    "array": None, # option for multi-task submission.
    "bb": None, "bbf": None, # burst buffer
    "begin": None, # scheduled runtime
    "chdir": None, # working directory of cluster process
    "clusters": None, # comma separated string of clusters
    "comment": None, # slurm script comment
    "constraint": None, # more constraints
    "deadline": None,
    "error": "job.err",
    "output": "job.out",
    "mail-user": None,
    "mail-type": None,
    # name / cpu / task / node allocation left to Manager
}

class MPRunner(object):
    """Interface for any kind of calculation that is run-able
    and set-able can be compatible with multiprocessing utilities like
    SlurmManager and DistributedPool."""
    def set_(self, **state_options):
        """Update the internal state of the object being run.

        Args:
            **state_options: Keywords arguments corresponding to attributes of the object
                being updated.
        """
        raise NotImplementedError("MPRunner must implement set_(self, **state_options) function.")

    def run(self, **run_options):
        """Run the calculation with the current settings.

        Args:
            **run_options: Keywords arguments that modify the nature of the
                way the program runs.
        """
        raise NotImplementedError("MPRunner must implement set_(self, **run_options) function.")

    def get_directory(self):
        """Return the directory in which the calculation is occurring.

        Returns
            (str): The path of the directory in which the program is being run.
        """
        raise NotImplementedError("MPRunner must implement get_directory(self) function.")

    def to_pickleable(self):
        """Returns a pickle-able portion of the object sufficient to run
        the calculations."""
        raise NotImplementedError("MPRunner must implement to_picklable(self) function.")

class DistributedPool(object):
    """A multiprocessing Pool context for running
    calculations among cores with different settings.

    Args:
        runner (MPRunner): The calculation runner.
        processes (int): The number of cores to divide up the work.
    """

    def __init__(self, runner: MPRunner, processes=None):
        # Always use the fork context
        mpc = mp.get_context("fork")
        self.runner = runner
        self.pool = mpc.Pool(processes=processes)

    def submit(self, run_args={}, **set_args):
        """Submit a single job with updated key words to the pool.

        Args:
            run_args (dict): Keyword arguments to be passed to
                :meth:`~.thunderboltz.parallel.MPRunner.run`.
        **set_args: Keyword arguments passed to
            :meth:`~.thunderboltz.parallel.MPRunner.set_`
            before calling :meth:`~.thunderboltz.parallel.MPRunner.run`.
        """
        # Copy the state of the runner (this ensures that
        # the parameter state does not get overwritten while
        # the processors are all busy)
        self.runner_child = copy.deepcopy(self.runner)
        # Keyword args will be sent to the set function
        self.runner_child.set_(**set_args)
        # run_args are sent separately with the caller
        self.pool.apply_async(
            self.runner_child.run, (), run_args,
            error_callback=self.err_callback)

    def __enter__(self):
        """Pipe the context input back to caller."""
        self.pool.__enter__()
        return self

    def err_callback(self, err):
        """Print out errors that subprocesses encounter."""
        path = self.runner.get_directory()
        warnings.warn(f"Error in process: PID={os.getpid()} "
                      f"path={path}\n\n{str(err)}", RuntimeWarning)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Safely wait for all processes to finish and exit the pool."""
        # Stop accepting jobs
        self.pool.close()
        # Wait for all jobs to finish
        self.pool.join()
        # Terminate multiprocessing pool
        self.pool.__exit__(exc_type, exc_val, exc_tb)

class SlurmManager(object):
    """A python context interface for the common Slurm HPC job manager
    to run more several intensive calculations on large clusters.
    See https://slurm.schedmd.com/sbatch.html.

    Args:
        runner (MPRunner): The calculation runner.
        directory (str or None) The simulation directory. Default is
            the current working directory.
        modules (list[str]): A list of modules to be loaded by the
            HPC module system.
        mock (bool): Option to test scripts without calling a slurm
            manager.
        **options: Additional keyword arguments will be interpreted as
            SLURM parameters.

    Note:
        This job manager currently only works for clusters that either
        already have the gcc and python requirements installed on each
        compute node, or clusters that use the
        `Module System <https://hpc-wiki.info/hpc/Modules>`_ to load
        functionality.
 
        The default behavior is to accommodate the module system as it
        is common on most HPC machines. If you wish to avoid writing
        ``module load`` commands in the SLURM script, simply specify
        ``modules=[]`` in the constructor.
    """
    _pckl_file = "job.pckl"
    def __init__(self, runner: MPRunner, directory=None, modules=["python", "gcc"],
            mock=False, **options):
        if directory is None:
            directory = os.getcwd()
        # Make group directory absolute
        if not os.path.isabs(directory):
            directory = str(Path(directory).resolve())
        #: The simulation directory
        self.directory = directory
        #: The list of modules to be loaded by the HPC module system.
        self.modules = modules
        #: The :class:`~.thunderboltz.parallel.MPRunner` object.
        self.runner = runner
        # Option to test the scripts without calling a slurm manager
        self.mock = mock
        # These will hold all the updates to both slurm and the runner
        self.input_sets = []
        #: Store references to the slurm job numbers after jobs are submitted
        self.job_ids = []
        #: The SLURM sbatch options
        self.options = slurm_options.copy()
        self.set_(**options)
        # Make sure cores are specified
        msg = (f"{self.__class__.__name__} requires "
                "the number of cores available at each node "
                "to be passed as ntasks-per-node.")
        if not "ntasks-per-node" in self.options:
            raise RuntimeError(msg)
        if not self.options["ntasks-per-node"]:
            raise RuntimeError(msg)

        # Assume fixed number of cores, store value
        self.ncores = self.options["ntasks-per-node"]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit and run sbatch for all the inputs"""
        ntasks = len(self.input_sets)
        nodes_required = int(np.ceil(ntasks / self.ncores))
        for start in range(nodes_required):
            # Isolate the inputs for this node
            inputs = self.input_sets[
                int(start*self.ncores):int(min((start+1)*self.ncores, ntasks))
            ]
            # Take any slurm updates from first input
            self.set_(**inputs[0])
            # Create a directory for slurm files of this node
            node_dir = pjoin(self.directory, f"NODE_{start+1}")
            os.makedirs(node_dir)
            with visit(node_dir):
                # Write pickle of input at the directory level
                with open(self._pckl_file, "wb") as f:
                    pickle.dump((inputs, self.runner), f)
                # Write the slurm script
                self.write_slurm_script()
                # Execute slurm call
                self.sbatch()

    def submit(self, run_args={}, **settings):
        """Add a set of parameter updates to the job queue.
        Slurm is not invoked until the context is exited.

        Args:
            run_args (dict): Keyword arguments to be passed to
                :meth:`~.thunderboltz.parallel.MPRunner.run`.
            **settings: Keyword arguments passed to the
                :meth:`~.thunderboltz.parallel.MPRunner.set_`.
                before calling :meth:`~.thunderboltz.parallel.MPRunner.run`.
        """
        # Resolve relative paths
        if not os.path.isabs(settings["directory"]):
            settings["directory"] = str(Path(settings["directory"]).resolve())

        combined_settings = {"run_args": run_args}
        combined_settings.update(**settings)
        self.input_sets.append(combined_settings.copy())

    def sbatch(self):
        """Call slurm with current settings."""
        if self.mock:
            # Just call the script and block
            self.mock_run()
            return
        # Otherwise, submit to slurm
        cmds = ["date | tee -a jobnum", "sbatch job_script.sh | tee -a jobnum"]
        subprocess.run(cmds[0], shell=True)
        sub_string = subprocess.check_output(cmds[1], shell=True).decode()
        self.job_ids.append(int(sub_string.split()[-1]))

    def set_(self, **options):
        """Update slurm manager options.

        Args:
            **options: SLURM settings.
        """
        self.options.update({k:v
            for k, v in options.items()
            if k in slurm_options})

    def write_slurm_script(self, path=None, script_name=None):
        """Write the SLURM batch script."""
        if script_name is None:
            script_name = "job_script.sh"
        if path is None:
            path = os.getcwd()
        with open(pjoin(path, script_name), "w") as f:
            f.write("#!/bin/bash\n")
            for k, v in self.options.items():
                if not v:
                    continue
                f.write(f"#SBATCH --{k}={v}\n")

            if not self.mock:
                # Only load modules if this is a real slurm run
                f.write("\n")
                for mod in self.modules:
                    f.write(f"module load {mod}\n")
                f.write("\n\n")

            python_script = self.process_batch_script()
            # Declare pyscript var
            f.write("SCRIPT=$(cat<<END\n")
            f.write(python_script)
            f.write("END\n")
            f.write(")\n\n")
            f.write("python -c \"$SCRIPT\"\n")

    def process_batch_script(self):
        """Inspect the batch script below and process it for use
        in sbatch."""
        source = inspect.getsource(self.batch_script)
        source = "\n".join(source.split("\n")[4:])
        # Inject ncores
        source = source.replace("ncores = None", f"ncores = {self.ncores}")
        # Inject pckl name
        source = source.replace("self._pckl_file", f"\"{self._pckl_file}\"")
        # Remove indents
        source = re.sub("^[^\S\n]{8}", "", source, flags=re.MULTILINE)
        return source

    def batch_script(self):
        """The SLURM job script. This does not get called in the
        parent process, but instead the source code is invoked in the
        sbatch script/command for subprocess startup."""
        import pickle
        from thunderboltz.parallel import DistributedPool
        from thunderboltz.parallel import slurm_options
        # number of cores and directory will be injected in sbatch
        ncores = None
        # Load in batch parameters
        with open(self._pckl_file, "rb") as f:
            updates, runner = pickle.load(f)
        # Run settings through distruted processing interface
        with DistributedPool(runner, processes=ncores) as pool:
            for update in updates:
                # Separate submit args
                submit_args = {k: v for k, v in update.items()
                    if k not in slurm_options}
                pool.submit(**submit_args)

    def mock_run(self):
        """Act as a compute node and test the job scripts sequentially."""
        os.system(f"chmod +x job_script.sh")
        os.system(f"./job_script.sh")

    def has_active(self):
        """Check whether any submitted jobs are still pending or running.

        Returns:
            (bool): ``True`` if there are still jobs that are pending or
                running. ``False`` otherwise.
        """
        has_active = False
        for job_id in self.job_ids:
            is_active = False
            try:
                cmd = f"scontrol show job {job_id}"
                sub_string = subprocess.check_output(
                    shlex.split(cmd)).decode()
                is_active = "PENDING" in sub_string or "RUNNING" in sub_string
            except subprocess.CalledProcessError as e:
                is_active = False
            has_active = has_active or is_active
        return has_active

    def has_pending(self):
        """Check whether any submitted jobs are still pending.

        Returns:
            (bool): ``True`` if there are still jobs that are pending.
                ``False`` otherwise.
        """
        has_pending = False
        for job_id in self.job_ids:
            is_active = False
            try:
                cmd = f"scontrol show job {job_id}"
                sub_string = subprocess.check_output(
                    shlex.split(cmd)).decode()
                is_pending = "PENDING" in sub_string
            except subprocess.CalledProcessError as e:
                is_pending = False
            has_pending = has_pending or is_pending
        return has_pending

    def join(self):
        """Wait for all slurm jobs to finish."""
        while self.has_active():
            time.sleep(2)

class visit(object):
    """Directory environment context manager."""
    def __init__(self, path):
        self.original_path = os.getcwd()
        self.path = path

    def __enter__(self):
        os.chdir(self.path)
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.original_path)

def setup_(i=0, subdir=None, base_directory=None):
    """Call from a run function such that it makes a directory
    in simulation/output with the name of that test function. Increment
    `i` if this function is being wrapped in another intermediary."""
    if base_directory is None:
        base_directory = os.getcwd()
    # Create the path to the test file
    fpath = pjoin(base_directory, inspect.stack()[1+i][3])
    if subdir is not None:
        fpath = pjoin(fpath, str(subdir))
    # Overwrite directory if necessary
    if os.path.isdir(fpath):
        shutil.rmtree(fpath)
    os.makedirs(fpath)
    return fpath




"""Plotting tool for intuitively comparing multidimensional data."""
from collections import OrderedDict
import copy
from functools import reduce
import itertools
import json
import os
from os.path import join as pjoin

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import InsetPosition
import numpy as np
from numpy import logical_and as land
import pandas as pd

# TEMP
from thunderboltz.plotting import styles

# Default color cycle
DCC = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd", "#8c564b",
        "#e377c2", "#7f7f7f", "#bcbd22", "#17becf", "#2fb", "#00F", "#F0F",
        "#039", "#00a", "#801"]

class PandaPlot(object):

    def __init__(self, df,
            x=None, y=None,
            err_bars={}, # Map from y key to errbar key
            label_map=lambda x: x, # Pretty column names
            value_map=lambda x: x, # Pretty parameter value names
            series=[],
            series_styles={},
            multi_legend=False,
            series_label_join=", ",
            series_label_interjoin="=",
            rows=None,
            cols=None,
            row_label_style="title", # title | annotation | None
            col_label_style="title", # title | annotation | None
            title_label_join=", ",
            suptitle=None,
            legend_style=None, # None | fig bottom | fig right
            broadcast_nans=True, # Series with missing ids are duplicated
            xscale="linear",
            yscale="linear",
            plot=True,
            gui_configure=False,
            save_spacing_config=None, # configuration file for plot spacing
            load_spacing_config=None, # configuration file for plot spacing
            config_dir=".", # place to store configuration files
            name="Figure", # figure name
            snapx=False,
            snapy=False,
            legend_ordering=None,
            remove_ext=False, # Remove extraneous labels.
            inset_bounds=None, # List of positions for the inset boxes
            **default_style,
            ):
        self.df = df
        self.x = x
        self.y = y
        self.err_bars = err_bars
        self.label_map = label_map
        self.value_map = value_map
        self.series = series
        self.series_styles = series_styles
        self.multi_legend = multi_legend
        self.series_label_join = series_label_join
        self.rows = rows
        self.cols = cols
        self.row_label_style = row_label_style
        self.col_label_style = col_label_style
        self.title_label_join = title_label_join
        self.suptitle = suptitle
        self.legend_style = legend_style
        self.broadcast_nans = broadcast_nans
        self.fig = None # Plot handles
        self.axs = None # Plot handles
        self.id_vars = [] # variables not bound to axis values
        self.xscale = xscale
        self.yscale = yscale
        self.default_style = default_style
        self.gui_configure = gui_configure
        self.save_spacing_config = save_spacing_config
        self.load_spacing_config = load_spacing_config
        self.config_dir = config_dir
        self.name = name
        self.snapx = snapx
        self.snapy = snapy
        self.remove_ext = remove_ext
        self.legend_ordering = legend_ordering
        self.inset_bounds=inset_bounds
        self.inset_axs = [[]] # store possible inset axes

        self.row_order = None
        self.col_order = None

        self.cfile = pjoin(config_dir, name)+".json"

        # err_bars should only include data in y
        if isinstance(self.y, list):
            self.err_bars = {k: v for k, v in self.err_bars.items() if k in self.y}
        else:
            self.err_bars = {k: v for k, v in self.err_bars.items() if k == self.y}

        self.validate_input()
        if plot:
            self.plot()

    def validate_input(self):
        # Generalize series into list
        if not isinstance(self.series, list):
            self.series = [self.series]
        # Store unique series keys in case of multi legend
        if self.multi_legend:
            series_keys = self.series
        else:
            series_keys = copy.deepcopy(self.series)
        if self.multi_legend:
            series_keys = []
            for k, v in self.series:
                series_keys.extend([k] + sum(v.values(), []))
        # Add style variables to series_keys
        series_keys.extend([k for k in self.df.columns if k in styles.presets])
        # Drop duplicates
        series_keys = list(set(series_keys))

        if isinstance(self.y, list):
            self.row_order = self.y
            # disallow additional figure row dim
            assert self.rows == None
            # melt y values into one df column
            id_vars = [self.x]+series_keys+[self.cols]
            id_vars = [idv for idv in id_vars if idv is not None]
            if id_vars:
                # Melt data with error bars
                df = pd.melt(self.df, id_vars=id_vars, value_vars=self.err_bars.keys(),
                    var_name="variable_y", value_name="value")
                df_err = pd.melt(self.df, id_vars=id_vars, value_vars=self.err_bars.values(),
                    var_name="variable_y", value_name="_error")
                df_werr = pd.concat((df, df_err[["_error"]]), axis=1)

                # Melt data without error bars
                nerr_names = [a for a in self.y if a not in self.err_bars.keys()]
                df_nerr = pd.melt(self.df, id_vars=id_vars, value_vars=nerr_names,
                    var_name="variable_y", value_name="value")

                # Combine data
                self.df = pd.concat((df_werr, df_nerr), ignore_index=False)

                # Reset y and rows
                self.y = "value"
                self.rows = "variable_y"
                self.err_bars = "_error"

        if isinstance(self.x, list):
            self.col_order = self.x
            # disallow additional figure column dim
            assert self.cols == None
            # melt x values into one df column
            id_vars = [self.x]+series_keys+[self.cols]
            id_vars = [idv for idv in id_vars if idv is not None]
            if id_vars:
                self.df = pd.melt(self.df, id_vars=id_vars, value_vars=self.y,
                    var_name="variable_x", value_name="value")
                # Reset x and cols
                self.x = "value"
                self.cols = "variable_x"

        # Determine variables not bound to axis values
        if self.rows not in ["variable_y", None]:
            self.id_vars += [self.rows]
        if self.cols not in ["variable_x", None]:
            self.id_vars += [self.cols]

        if self.broadcast_nans:
            # Check if rows y-values
            for idv in self.id_vars:
                # Select data with and without nans.
                missing = self.df[self.df[idv].isna()].copy()
                not_missing = self.df.loc[~self.df[idv].isna()].copy()
                col_names = not_missing[idv].unique()
                self.df = not_missing.copy()
                for col in col_names:
                    missing[idv] = col
                    self.df = pd.concat((self.df, missing), ignore_index=True)

    def plot(self):
        df = self.df
        # Determine number of panels
        nrows = ncols = 1
        if self.rows:
            nrows = len(df[self.rows].unique())
        if self.cols:
            ncols = len(df[self.cols].unique())
        # TODO: autogenerate figsize more accurately
        self.fig, self.axs = plt.subplots(nrows, ncols, figsize=(5*nrows,
            5*ncols), squeeze=False)
        self.inset_axs = [[[] for j in range(ncols)] for i in range(nrows)]

        if self.multi_legend:
            multi_leg_data = self.build_multi_legend(df)
        else:
            # Set a default cycle for unstyled series
            defc = itertools.cycle(DCC)
            for s in df.series.unique():
                if s in self.series_styles:
                    if "c" not in self.series_styles[s]:
                        self.series_styles[s]["c"] = next(defc)
                else:
                    self.series_styles[s] = {"c": next(defc)}
        # Loop though panels and series
        for i, (row_name, rdf) in enumerate(groupby_iter(df, self.rows, order=self.row_order)):
            for j, (col_name, cdf) in enumerate(groupby_iter(rdf, self.cols, order=self.col_order)):
                # Select panel
                ax = self.axs[i, j]
                if self.inset_bounds is not None:
                    self.add_inset(i, j, ax)
                if self.multi_legend:
                    for handle, label, style in zip(*multi_leg_data):
                        sdf = cdf[require(cdf, handle)].copy()
                        self.plot_simple(i, j, ax, sdf, row_name, col_name, label, style)
                else:
                    for s_name, sdf in groupby_iter(cdf, self.series):
                        s_str = self.series_label_join.join(s_name)
                        # Style for this series
                        style = copy.deepcopy(self.default_style)
                        if s_str in self.series_styles:
                            style.update(self.series_styles[s_str])
                        label = dict_f(self.label_map, s_str)
                        self.plot_simple(i, j, ax, sdf, row_name, col_name, label, style)

        if self.suptitle:
            self.fig.suptitle(self.suptitle)

        if self.remove_ext:
            remove_ext_labels(self.axs)

        fs = 18
        if self.multi_legend:
            fs = 11
        if self.legend_style == "figure bottom":
            handles, labels = figure_legend(self.axs.flatten(), order=self.legend_ordering)
            self.figure_legend = self.fig.legend(handles, labels, loc="lower center",
                    ncol=1, prop={"size": fs, "family": "monospace"})
        if self.legend_style == "figure right":
            handles, labels = figure_legend(self.axs.flatten(), order=self.legend_ordering)
            self.figure_legend = self.fig.legend(handles, labels, loc="center right", ncol=1,
                    prop={"size": fs, "family": "monospace"})
            self.fig.subplots_adjust(right=0.6)

        if self.gui_configure:
            # Try to load a config if it exists.
            if os.path.exists(self.cfile) and self.load_spacing_config:
                self.read_spacing_config(self.cfile)

            plt.show()

            self.snap_lims()

            if self.save_spacing_config:
                self.write_spacing_config(self.cfile)

        if self.load_spacing_config:
            self.read_spacing_config(self.cfile)

        return self.fig, self.axs

    def plot_simple(self, i, j, ax, dat, row_name, col_name, label, style):
        """Plot a single set of data with a fixed style"""
        # if dat[self.x].isna().all():
            # return
        # Order data
        dat = dat.sort_values(self.x).reset_index(drop=True)
        # Plot everything
        if dat.isna().all().all():
            line = ax.plot([], [], label=label, **style)
        else:
            line = ax.plot(dat[self.x], dat[self.y], label=label, **style)

        # Optionally add err bars.
        if self.err_bars and not dat[self.err_bars].isna().all():
            ax.errorbar(dat[self.x], dat[self.y], dat[self.err_bars],
                    ls="", ecolor=line[0].get_color(),
                    elinewidth=1, capsize=2)

        # Do the same for insets, without forming legend
        if self.inset_axs[i][j] and not dat.isna().all().all():
            self.inset_axs[i][j].plot(dat[self.x], dat[self.y], **style)
            if self.err_bars and not dat[self.err_bars].isna().all():
                ax.errorbar(dat[self.x], dat[self.y], dat[self.err_bars],
                        ls="", ecolor=line[0].get_color(), elinewidth=1, capsize=2)
        # Label axis
        ax.set_xlabel(dict_f(self.label_map, self.x))
        ax.set_ylabel(dict_f(self.label_map, self.y))
        # Title panels
        titles = []
        if self.row_label_style=="title" and self.rows != "variable_y":
            titles.append(row_name)
        if self.col_label_style=="title" and self.cols != "variable_x":
            titles.append(col_name)
        ax.set_title(self.title_label_join.join(list(map(str,
            titles))), fontsize=16)
        if self.x == "value":
            ax.set_xlabel(dict_f(self.label_map, col_name))
        if self.y == "value":
            ax.set_ylabel(dict_f(self.label_map, row_name))
        # Configure axes scales
        ax.set_xscale(self.xscale)
        ax.set_yscale(self.yscale)
        ax.grid(which='major', color='gray', linestyle='-', alpha=0.5)
        ax.grid(which='minor', color='gray', linestyle='-', alpha=0.3)
        if self.inset_axs[i][j]:
            self.inset_axs[i][j].set_xscale(self.xscale)
            # self.inset_axs[-1].set_yscale(self.yscale)
            self.inset_axs[i][j].grid(which='major', color='gray', linestyle='-', alpha=0.5)
            self.inset_axs[i][j].grid(which='minor', color='gray', linestyle='-', alpha=0.3)
        plt.minorticks_on()
        if self.legend_style is None:
            # Create legend
            ax.legend(fontsize=14, ncol=2)

    def add_inset(self, i, j, ax):
        if self.inset_bounds[i][j] is None:
            self.inset_axs[i][j] = None
        else:
            inset_ax = plt.axes([0, 0, 1, 1])
            inset_ax.set_axes_locator(InsetPosition(ax, self.inset_bounds[i][j]))
            self.inset_axs[i][j] = inset_ax

    def write_spacing_config(self, cfile):
        fs = tuple(self.fig.get_size_inches())
        lims = [[ax.get_xlim(), ax.get_ylim()] for ax in self.axs.flatten()]
        spps = {p: getattr(self.fig.subplotpars, p) for p in
                ["left", "right", "top", "bottom", "wspace", "hspace"]}
        dat = {"fs": fs, "lims": lims, "spps": spps}
        with open(cfile, "w") as f:
            f.write(json.dumps(dat))

    def read_spacing_config(self, cfile):
        with open(cfile, "r") as f:
            d = json.load(f)
        self.fig.set_size_inches(d["fs"])
        self.fig.subplots_adjust(**d["spps"])
        for ax, lims in zip(self.axs.flatten(), d["lims"]):
            ax.set_xlim(lims[0])
            ax.set_ylim(lims[1])

    def snap_lims(self):
        lims = np.array([[list(ax.get_xlim())+list(ax.get_ylim())
                          for ax in axrow] for axrow in self.axs])
        if self.snapx:
            lims[:,:,0] = np.min(lims[:,:,0], axis=0)
            lims[:,:,1] = np.max(lims[:,:,1], axis=0)
        if self.snapy:
            mins = np.min(lims[:,:,2], axis=1).reshape(-1, 1)
            maxs = np.max(lims[:,:,3], axis=1).reshape(-1, 1)
            lims[:,:,2] = np.repeat(mins, 2, axis=1)
            lims[:,:,3] = np.repeat(maxs, 2, axis=1)

        for axrow, limrow in zip(self.axs, lims):
            for ax, lims in zip(axrow, limrow):
                ax.set_xlim(lims[:2])
                ax.set_ylim(lims[2:])

    def save(self, fdir):
        self.fig.savefig(pjoin(fdir, self.name+".pdf"))

    def build_multi_legend(self, df):
        """Create legend data from DataFrame of raw data, legend table format,
        column label maps, value label maps, and style maps."""
        # find preset keys that should not be included in legend
        pset_keys = [k for k in df if k in styles.presets and k not in [k_[0] for
                     k_ in self.series]]
        # Create style cycles, if they aren't cycles already
        for k in self.series_styles:
            if not isinstance(self.series_styles[k][2], itertools.cycle):
                self.series_styles[k][2] = itertools.cycle(self.series_styles[k][2])
        dfs = []
        # Isolate descriptor variables
        for key, col_d in self.series:
            for val, cols in col_d.items():
                dfs.append(df[df[key] == val][[key]+cols+pset_keys].drop_duplicates(
                    ).reset_index(drop=True))
            # Add everything else
            df_sep = df[~df[key].isin(col_d.keys())][[key]+pset_keys].drop_duplicates(
                ).reset_index(drop=True)
            # If there is anything else
            if len(df_sep):
                dfs.append(df_sep)
        # Store variable handles
        hands = []
        # Store legend strings
        legs = []
        # Store the styles that correspond to each legend strings
        stys = []
        # Loop through tables of legend
        for df_ in dfs:
            df_ = df_.sort_values(list(df_.columns)).reset_index(drop=True)
            # Make dataframe with string data types
            sdf = df_.copy()
            for col in sdf:
                sdf[col] = sdf[col].map(lambda v: self.value_map(col)(v)).astype("str")
            sdf.drop(pset_keys, axis=1, inplace=True)
            # Convert dfs to strings
            formatters = {}
            for col in sdf.columns:
                len_max = sdf[col].str.len().max()
                # Ignore special math operators that won't show up on the plot
                len_max = max(len_max, len(str(self.label_map(col).replace("$", "").replace("^", ""))))
                formatters[col] = len_max
            # Add legend header and blank legend header linestyle
            hands += [{}]
            legs += [" ".join(f"{self.label_map(s):^{mx}s}" for s, mx in formatters.items())]
            stys += [dict(ls="", marker="")] # Add empty style
            for (_, row), (_, srow) in zip(df_.iterrows(), sdf.iterrows()):
                # Add style for this line using actual data
                stys += [{}]
                # First check if there are specified styles from `self.series_styles`
                for col, (vs, styk, styvs) in self.series_styles.items():
                    if vs == None:
                        vs = [row[col]]
                        styvs = styles.DCYC[styk]
                    for v, styv in zip(vs, styvs):
                        if col in row and row[col] == v:
                            stys[-1].update({styk: styv})

                # Then check if there are any preset styles for this row
                for pskey in styles.presets:
                    for psval, pssty in styles.presets[pskey].items():
                        if pskey in row.index and row[pskey] == psval:
                            stys[-1].update(
                                {k: v for k, v in next(pssty).items()
                                if k not in stys[-1]}
                            )
                # Drop pset keys from row
                row = row[[k for k in row.index if k not in pset_keys]].copy()
                # Add legend text to legend list using string data.
                legs += [" ".join(f"{s:<{mx}s}" for col, s, mx in
                    zip(srow.index, srow.values, formatters.values()))]
                # Add data from `row` into handles list
                hands += [row.to_dict()]

        return hands, legs, stys

# Util
def get_hashable(h):
    """Return a hashable string that uniquely
    describes the plotting style of a handle."""
    props = [h.get_ls(), h.get_lw(), h.get_c(),
        h.get_ms(), h.get_marker()]
    s = "".join(list(map(str, props)))
    return s

def figure_legend(axs, order=None):
    """Take several axis and merge their legends into one
    figure legend."""
    # Store unique handle representations.
    hs = set()
    # Map back to the actual line objects.
    lhmap = OrderedDict()
    for ax in axs:
        hand, lab = ax.get_legend_handles_labels()
        lm = {get_hashable(h)+"@"+l: h for l, h in zip(lab, hand)}
        lhmap.update(lm)
    labels = [l.split("@")[1] for l in lhmap.keys()]
    handles = list(lhmap.values())
    if order == None:
        return handles, labels
    return [handles[i] for i in order], [labels[i] for i in order]

def dict_f(d, x, default=None):
    if isinstance(d, dict):
        if x in d:
            return d[x]
        elif default is not None:
            return default
        return x
    return d(x)

def groupby_iter(df, groups, order=None):
    if groups == None:
        yield "", df
    elif order == None:
        for name, sub_df in df.groupby(groups):
            yield name, sub_df
    else:
        d = {name: sub_df for name, sub_df in df.groupby(groups)}
        for name in order:
            yield name, d[name]

def remove_ext_labels(axs_):
    # Determine shape of the figure.
    ncols = axs_.shape[1]
    axs = axs_.flatten()
    n_last_row = ncols
    nrows = len(axs) // ncols
    if len(axs) % ncols:
        n_last_row = len(axs) % ncols
        nrows = ((len(axs) - n_last_row) // ncols) + 1
    def split_set(n):
        # Return the number(s) half way to n
        if not n % 2:
            return {n // 2, n // 2 + 1}
        return {(n // 2) + 1}
    axl = [axs[n*ncols: (n+1)*ncols] for n in range(nrows - 1)] + [axs[-n_last_row:]]
    for i, row in enumerate(axl):
        for j, ax in enumerate(row):
            if not i+1 in split_set(nrows) or j:
                axl[i][j].set_ylabel("")
            if not j+1 in split_set(n_last_row) or nrows-i-1:
                axl[i][j].set_xlabel("")

def require(df, d, other=[]):
    """Return a mask where all the columns have values passed by the dictionary."""
    if d == {}:
        return pd.DataFrame(None, columns=df.columns)
    crit = [df[k] == v for k, v in d.items()]
    return reduce(land, crit+other)




"""Labels, formats, plot styles, etc."""
import itertools

ls_dashed1 = (0, (1.05, 1.6, 6.5, 1.6))
ls_dashed2 = (0, (4, 2))
ls_dots1 = (0, (2.05, 1.5, 1.5, 3.05, 2.05, 1.42))
ls_dots2 = (0, (1.65, 3.5, 3.05, 2.05))
ls4 = (0, (3.5, 3.5))
ls10 = (0, (.4, 1.1, .4, 1.1, .4, 2))

# Default color cycle
DCC = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd", "#8c564b",
        "#e377c2", "#7f7f7f", "#bcbd22", "#17becf", "#2fb", "#00F", "#F0F",
        "#039", "#00a", "#801"]


# General default cycles
DCYC = {
    "ls": [ls_dashed1, ls_dashed2, ls_dots1, ls_dots2, ls4, ls10],
    "c": DCC,
    "marker": [".", "o", "^", "3", "1"],
}

# Prefabricated style cycles
presets = {
    "data_class": {
        "experimental": itertools.cycle({"markersize": 6, "marker": "o", "ls": "",
                        "alpha": 0.9, "c": c} for c in ["red", "green",
                        "orange", "blue", "#afa", "#aa0852", "#dba", "#abd"])},
    "name": {
        "Bolsig": itertools.cycle({"ls": "--", "c": c, "marker": None, "fillstyle": "none"} for c in DCC),
        "ThunderBoltz": itertools.cycle({"ls": "", "marker": "v", "c": c, "markersize": 12} for c in DCC),
        "Aleph": itertools.cycle([{"ls": "", "marker": "v", "c": "black", "markersize": 9}])},
}


def label_maps(s):
    """Mappings for neat math representations of parameters and column names."""
    # Non math text
    text = {
        "name": "",
        "eadf": "EADF",
        "eesd": "EESDF",
        "egen": "e$^{-}$ gen",
        "pc_scale": "Scaled Param",
        "kucukarpaci et al 1981.": 'K' + u'\u00FC' + 'c' + u'\u00FC'  + r"karpaci $et$ $al.$ 1981",
        # "el_rot": "elastic Euler rot",
        # "ion_rot": "ionizaiton Euler rot",
    }
    if s in text:
        return text[s]

    if "." in s:
        # Assume this is an experiment, remove the period at the end.
        s = s[:-1]


    if "et al" in s:
        return s.replace("et al", "$et$ $al.$")

    # Base mappings
    math_text = { # Symbol, unit symbol
        "a_n": r"\alpha/n_{\rm gas}\rm{\ (m}^{2}\rm{)}",
        "a_n_-21": r"\alpha/n_{\rm gas}\rm{\ (10}^{-21}{\rm m}^{2}\rm{)}",
        "a_n_bulk": r"\alpha/n_{\rm gas}\rm{\ \ bulk}",
        "a_n_bulk_fit": r"\alpha/N\rm{\ \ bulk}",
        "MEe": r"\langle E \rangle",
        "DT": "\Delta t",
        "DE": r"\Delta \epsilon",
        "mobN": r"\mu n_{\rm gas}\rm{\ (10}^{24}\rm{V}^{-1}\rm{m}^{-1}\rm{s}^{-1}\rm{)}",
        "mobN_bulk": r"\mu n_{\rm gas}\rm{\ (10}^{24}\rm{V}^{-1}\rm{m}^{-1}\rm{s}^{-1}\rm{)\ bulk}",
        "mobN_bulk_fit": r"\mu n_{\rm gas}\rm{\ (10}^{24}\rm{V}^{-1}\rm{m}^{-1}\rm{s}^{-1}\rm{)\ bulk}",
        "L": r"\rm{L (m)}",
        "Ered": r"\rm{E/n_{\rm gas}\ (Td)}",
        "k_tot": r"\rm{k}_{tot}",
        "name": r"\rm{}",
        "k": r"\rm{k\ (m}^3\rm{/s)}",
    }
    # Ignore if no label
    if s.replace("_std", "").replace("_err", "") not in math_text.keys():
        return s


    def alter(s):
        news, typ = "_".join(s.split("_")[:-1]), s.split("_")[-1]
        if typ == "std":
            return r"\sigma_{" + alter(news) + "}"
        elif typ == "err":
            alt = r"\%\ \rm{err}\ " + alter(news)
            # Remove units
            alt = re.sub("\(.+\)", "", alt)
            return alt
        elif s in math_text:
            # Base case, no alterations, just use label mapping
            return math_text[s]

    # Alterations such as standard deviation and error
    # return alter(s)
    return f"${alter(s)}$"

def value_maps(key):
    vmaps = {
        "DT": lambda x: f"{x:.2e}",
        "growth": lambda x: {2.0: "SST", 1.0: "PT", 4.0: "GE"}[x],
        "pc_scale": lambda x: "None" if not len(x) else str(x)
    }

    if key in vmaps:
        return vmaps[key]
    return lambda x: x
